package ava.parser

import cats.effect.IO
import cats.effect.Resource
import cats.effect.unsafe.IORuntime
import java.io.ByteArrayInputStream
import java.io.InputStream

class TokenizerTests extends munit.FunSuite:
  implicit val runtime: IORuntime = cats.effect.unsafe.IORuntime.global

  private val EmptyStream: InputStream = new ByteArrayInputStream(Array[Byte]())

  test("should produce the EOF token for an empty stream") {
    run(newCR(EmptyStream)) { tokenizer =>
      IO(assertEquals(tokenizer.next(), Right(Token.Eof)))
    }
  }

  test(
    "should produce the EOF token for a stream of only whitespace characters"
  ) {
    run(newCR(stringStream(" \t\n\r\n   "))) { tokenizer =>
      IO(assertEquals(tokenizer.next(), Right(Token.Eof)))
    }
  }

  test(
    "should produce an error, followed by EOF, if only a backslash chararacter is present"
  ) {
    assertTokens(
      " \\ ",
      Left(Tokenizer.Error.BackSlashNotAllowed(pos(2, 1, 2)))
    )
  }

  test("should capture all fixed, single-character tokens") {
    assertTokens(
      "(),:.#",
      Right(Token.OpenParen),
      Right(Token.CloseParen),
      Right(Token.Comma),
      Right(Token.Colon),
      Right(Token.Dot),
      Right(Token.Tuple)
    )
  }

  test("should not recognize comments if preceded by any non-whitespace") {
    assertTokens(
      "( --",
      Right(Token.OpenParen),
      Right(Token.Generic("--"))
    )
  }

  test("should recognize potential comments that are just a dash (space)") {
    assertTokens("- ", Right(Token.Generic("-")))
  }

  test("should recognize potential comments that are just a dash (tab)") {
    assertTokens("-\t", Right(Token.Generic("-")))
  }

  test("should recognize potential comments that are just a dash (newline)") {
    assertTokens("-\n", Right(Token.Generic("-")))
  }

  private def assertTokens(
    source: String,
    expectedOutput: Either[Tokenizer.Error, Token]*
  ): Unit =
    run(newCR(stringStream(source))) { tokenizer =>
      consumeAll(tokenizer).map { tokens =>
        assertEquals(tokens, expectedOutput.toList)
      }
    }

  private def consumeAll(tokenizer: Tokenizer)
    : IO[List[Either[Tokenizer.Error, Token]]] =
    fs2.Stream
      .repeatEval(IO(tokenizer.next()))
      .takeWhile(_ != Right(Token.Eof))
      .compile
      .toList

  private def pos(
    charInFile: Long,
    lineNumber: Long,
    charInLine: Long
  ): SourcePosition =
    SourcePosition(charInFile, lineNumber, charInLine)

  private def stringStream(contents: String): InputStream =
    new ByteArrayInputStream(contents.getBytes())

  private def newCR(is: InputStream): CharacterReader =
    CharacterReader.forInputStream(is)

  private def run(reader: CharacterReader)(testF: Tokenizer => IO[Unit]): Unit =
    Resource
      .make(
        acquire = IO(new Tokenizer(reader))
      )(release = t => IO(t.close()))
      .use(testF)
      .unsafeRunSync()
